@@ -1,15 +1,18 @@
 package HTTP::Server::Simple;
-use 5.006;
 use strict;
 use warnings;
+use FileHandle;
 use Socket;
 use Carp;
+use URI::Escape;
+
+use vars qw($VERSION $bad_request_doc);
+$VERSION = '0.28_01';
 
-our $VERSION = '0.14_03';
 
 =head1 NAME
 
-HTTP::Server::Simple
+HTTP::Server::Simple - Lightweight HTTP server
 
 
 =head1 SYNOPSIS
@@ -55,7 +58,12 @@ By default, the server traps a few signals:
 
 =item HUP
 
-When you C<kill -HUP> the server, it does its best to rexec itself.
+When you C<kill -HUP> the server, it lets the current request finish being
+processed, then uses the C<restart> method to re-exec itself. Please note that
+in order to provide restart-on-SIGHUP, HTTP::Server::Simple sets a SIGHUP
+handler during initialisation. If your request handling code forks you need to
+make sure you reset this or unexpected things will happen if somebody sends a
+HUP to all running processes spawned by your app (e.g. by "kill -HUP <script>")
 
 =item PIPE
 
@@ -71,6 +79,7 @@ API call to start a new server.  Does not actually start listening
 until you call C<-E<gt>run()>.
 
 =cut
+
 sub new {
     my ( $proto, $port ) = @_;
     my $class = ref($proto) || $proto;
@@ -83,9 +92,35 @@ sub new {
     my $self = {};
     bless( $self, $class );
     $self->port( $port || '8080' );
+
+
+
     return $self;
 }
 
+
+=head2 lookup_localhost
+
+Looks up the local host's hostname and IP address.
+
+Stuffs them into
+
+$self->{'localname'} and $self->{'localaddr'}
+
+=cut
+
+sub lookup_localhost {
+    my $self = shift;
+
+    my $local_sockaddr = getsockname( $self->stdio_handle );
+    my ( undef, $localiaddr ) = sockaddr_in($local_sockaddr);
+    $self->host( gethostbyaddr( $localiaddr, AF_INET ) || "localhost");
+    $self->{'local_addr'} = inet_ntoa($localiaddr) || "127.0.0.1";
+}
+
+
+
+
 =head2 port [NUMBER]
 
 Takes an optional port number for this server to listen on.
@@ -128,9 +163,9 @@ sub background {
     my $child = fork;
     die "Can't fork: $!" unless defined($child);
     return $child if $child;
-    use POSIX;
 
     if ( $^O !~ /MSWin32/ ) {
+        require POSIX;
         POSIX::setsid()
             or die "Can't start a new session: $!";
     }
@@ -146,31 +181,14 @@ start listening for http requests.
 
 my $server_class_id = 0;
 
+use vars '$SERVER_SHOULD_RUN';
+$SERVER_SHOULD_RUN = 1;
+
 sub run {
     my $self   = shift;
     my $server = $self->net_server;
 
-    # Handle SIGHUP
-
     local $SIG{CHLD} = 'IGNORE';    # reap child processes
-    local $SIG{HUP} = sub {
-        close HTTPDaemon;
-
-        # and then, on systems implementing fork(), we make sure
-        # we are running with a new pid, so another -HUP will still
-        # work on the new process.
-        require Config;
-        if ( $Config::Config{d_fork} and my $pid = fork() ) {
-
-            # finally, allow ^C on the parent process to terminate
-            # the children.
-            waitpid( $pid, 0 );
-            exit;
-        }
-
-        # do the exec. if $0 is not executable, try running it with $^X.
-        exec {$0}( ( ( -x $0 ) ? () : ($^X) ), $0, @ARGV );
-    };
 
     # $pkg is generated anew for each invocation to "run"
     # Just so we can use different net_server() implementations
@@ -187,9 +205,12 @@ sub run {
     }
     else {
         $self->setup_listener;
+	$self->after_setup_listener();
         *{"$pkg\::run"} = $self->_default_run;
     }
 
+    local $SIG{HUP} = sub { $SERVER_SHOULD_RUN = 0; };
+
     $pkg->run( port => $self->port );
 }
 
@@ -212,14 +233,16 @@ sub _default_run {
 
         $self->print_banner;
 
-        while (1) {
+        while ($SERVER_SHOULD_RUN) {
             local $SIG{PIPE} = 'IGNORE';    # If we don't ignore SIGPIPE, a
                  # client closing the connection before we
                  # finish sending will cause the server to exit
-            while ( accept( my $remote, HTTPDaemon ) ) {
+            while ( accept( my $remote = new FileHandle, HTTPDaemon ) ) {
                 $self->stdio_handle($remote);
+                $self->lookup_localhost() unless ($self->host);
                 $self->accept_hook if $self->can("accept_hook");
 
+
                 *STDIN  = $self->stdin_handle();
                 *STDOUT = $self->stdout_handle();
                 select STDOUT;   # required for HTTP::Server::Simple::Recorder
@@ -228,9 +251,38 @@ sub _default_run {
                 close $remote;
             }
         }
+
+        # Got here? Time to restart, due to SIGHUP
+        $self->restart;
     };
 }
 
+=head2 restart
+
+Restarts the server. Usually called by a HUP signal, not directly.
+
+=cut
+
+sub restart {
+    my $self = shift;
+
+    close HTTPDaemon;
+
+    $SIG{CHLD} = 'DEFAULT';
+    wait;
+
+    ### if the standalone server was invoked with perl -I .. we will loose
+    ### those include dirs upon re-exec. So add them to PERL5LIB, so they
+    ### are available again for the exec'ed process --kane
+    use Config;
+    $ENV{PERL5LIB} .= join $Config{path_sep}, @INC;
+
+    # Server simple
+    # do the exec. if $0 is not executable, try running it with $^X.
+    exec {$0}( ( ( -x $0 ) ? () : ($^X) ), $0, @ARGV );
+}
+
+
 sub _process_request {
     my $self = shift;
 
@@ -245,40 +297,35 @@ sub _process_request {
         binmode STDIN,  ':raw';
         binmode STDOUT, ':raw';
 
+        # The ternary operator below is to protect against a crash caused by IE
+        # Ported from Catalyst::Engine::HTTP (Originally by Jasper Krogh and Peter Edwards)
+        # ( http://dev.catalyst.perl.org/changeset/5195, 5221 )
+        
         my $remote_sockaddr = getpeername( $self->stdio_handle );
-        my ( undef, $iaddr ) = sockaddr_in($remote_sockaddr);
-        my $peername = gethostbyaddr( $iaddr, AF_INET ) || "localhost";
-
-        my $peeraddr = inet_ntoa($iaddr) || "127.0.0.1";
-
-        my $local_sockaddr = getsockname( $self->stdio_handle );
-        my ( undef, $localiaddr ) = sockaddr_in($local_sockaddr);
-        my $localname = gethostbyaddr( $localiaddr, AF_INET )
-            || "localhost";
-        my $localaddr = inet_ntoa($localiaddr) || "127.0.0.1";
-
-        my ( $method, $request_uri, $proto ) = $self->parse_request
-            or do { $self->bad_request; return };
-
-        $proto ||= "HTTP/0.9";
-
-        my ( $file, $query_string )
-            = ( $request_uri =~ /([^?]*)(?:\?(.*))?/ );    # split at ?
-
-        if ( $method !~ /^(?:GET|POST|HEAD)$/ ) {
+        my ( undef, $iaddr ) = $remote_sockaddr ? sockaddr_in($remote_sockaddr) : (undef,undef);
+        my $peeraddr = $iaddr ? ( inet_ntoa($iaddr) || "127.0.0.1" ) : '127.0.0.1';
+        
+        my ( $method, $request_uri, $proto ) = $self->parse_request;
+        
+        unless ($self->valid_http_method($method) ) {
             $self->bad_request;
             return;
         }
 
+        $proto ||= "HTTP/0.9";
+
+        my ( $file, $query_string )
+            = ( $request_uri =~ /([^?]*)(?:\?(.*))?/s );    # split at ?
+
         $self->setup(
             method       => $method,
             protocol     => $proto,
-            query_string => ( $query_string || '' ),
+            query_string => ( defined($query_string) ? $query_string : '' ),
             request_uri  => $request_uri,
             path         => $file,
-            localname    => $localname,
+            localname    => $self->host,
             localport    => $self->port,
-            peername     => $peername,
+            peername     => $peeraddr,
             peeraddr     => $peeraddr,
         );
 
@@ -298,6 +345,10 @@ sub _process_request {
         }
 }
 
+
+
+
+
 =head2 stdio_handle [FILEHANDLE]
 
 When called with an argument, sets the socket to the server to that arg.
@@ -387,7 +438,7 @@ of keys of this list.
 =cut
 
 sub setup {
-    my ($self) = @_;
+    my $self = shift;
     while ( my ( $item, $value ) = splice @_, 0, 2 ) {
         $self->$item($value) if $self->can($item);
     }
@@ -552,6 +603,16 @@ sub setup_listener {
 
 }
 
+
+=head2 after_setup_listener
+
+This method is called immediately after setup_listener. It's here just for you to override.
+
+=cut
+
+sub after_setup_listener {
+}
+
 =head2 bad_request
 
 This method should print a valid HTTP response that says that the
@@ -559,7 +620,7 @@ request was invalid.
 
 =cut
 
-our $bad_request_doc = join "", <DATA>;
+$bad_request_doc = join "", <DATA>;
 
 sub bad_request {
     my $self = shift;
@@ -569,9 +630,22 @@ sub bad_request {
         length($bad_request_doc), "\r\n\r\n", $bad_request_doc;
 }
 
+=head2 valid_http_method($method)
+
+Given a candidate HTTP method in $method, determine if it is valid.
+Override if, for example, you'd like to do some WebDAV.
+
+=cut 
+
+sub valid_http_method {
+    my $self   = shift;
+    my $method = shift or return 0;
+    return $method =~ /^(?:GET|POST|HEAD|PUT|DELETE)$/;
+}
+
 =head1 AUTHOR
 
-Copyright (c) 2004-2005 Jesse Vincent, <jesse@bestpractical.com>.
+Copyright (c) 2004-2006 Jesse Vincent, <jesse@bestpractical.com>.
 All rights reserved.
 
 Marcus Ramberg <drave@thefeed.no> contributed tests, cleanup, etc
